# coding: utf-8
import redis
from urllib.parse import urlparse
import logging

logger = logging.getLogger(__name__)


def url_change(redis_url):
    url = urlparse(redis_url)
    host = url.hostname, port = url.port, password = url.password
    return host, port, password


class RedisObj:
    def __init__(self, host, port, db, decode_responses=True):
        try:
            self.obj = redis.StrictRedis(
                host=host,
                port=port,
                db=db,
                decode_responses=decode_responses
            )
            # name = 'task_manage'
            # 先校验是否成功连接数据库
            # self.list_all(name)
        except redis.exceptions.ConnectionError as e:
            logger.error('redis 数据库连接失败')
            exit()

    @property
    def redis_obj(self):
        ''' 返回redis对象 '''
        return self.obj

    def length(self, name):
        ''' 获取list的长度 '''
        return self.obj.llen(name)

    def push(self, name, *value):
        ''' 默认为向list后面插入数据 '''
        return self.obj.rpush(name, *value)

    def pop(self, name):
        ''' 默认从list前面弹出数据 '''
        # return self.obj.brpop(name, timeout=self.timeout)
        return self.obj.lpop(name)

    def list_all(self, name):
        ''' 查看list的所有数据 '''
        return self.obj.lrange(name, 0, -1)

    def task_clear(self, name):
        ''' 清空指定的列表内元素 1表示已移除，0表示不存在'''
        return self.obj.delete(name)

    def task_list(self):
        ''' 查看所有任务列表 '''
        return self.obj.keys()

    def task_exists(self, task_id):
        ''' 判断任务是否存在 1为存在 0为不存在'''
        return self.obj.exists(task_id)



"""
Example of simple consumer that waits for a single message, acknowledges it
and exits.
"""
import logging
from pprint import pformat
from threading import Thread
from kombu import Connection, Consumer, Exchange, Queue, eventloop, Producer
from kombu.pools import producers
from amqp.exceptions import ConnectionForced

from ..file import read_yaml

logger = logging.getLogger(__name__)


# priority_to_routing_key = {
#     'high': 'hipri',
#     'mid': 'midpri',
#     'low': 'lopri',
# }


def pretty(obj):
    return pformat(obj, indent=4)


class Broker:
    queue_attrs = ['queue_arguments', 'binding_arguments', 'consumer_arguments', 'durable', 'exclusive', 'auto_delete',
                   'no_ack', 'alias', 'bindings', 'no_declare', 'expires', 'message_ttl', 'max_length',
                   'max_length_bytes', 'max_priority']

    exchange_attrs = ['arguments', 'durable', 'passive', 'auto_delete', 'delivery_mode', 'no_declare']

    def __init__(self, broker_ip='localhost', port=5672, username=None, password=None, queue=None, routing_key=None,
                 exchange=None, exchange_type=None, **kwargs):
        self.exchanges = {}
        self.queues = {}
        self.host = broker_ip
        self.port = port
        self.username = username
        self.password = password
        self.vhost = kwargs['vhost'] if 'vhost' in kwargs else '%2F'
        self.kwargs = kwargs
        if queue and routing_key and exchange_type and exchange:
            _exchange_attr = self._find_attr(self.exchange_attrs, self.kwargs)
            _queue_attr = self._find_attr(self.queue_attrs, self.kwargs)
            self.queue, self.exchange = self.add_conn(queue, routing_key, exchange_type, exchange, _queue_attr,
                                                      _exchange_attr)

        self._conn = None

    def load(self, config_dict=None, config_path=None, enforce=True):
        config = {}
        if config_path:
            try:
                config.update(read_yaml(config_path, 'MQCONFIG'))
            except Exception as e:
                logger.error(e, exc_info=True)
                return
        elif config_dict:
            config.update(config_dict)
        else:
            return

        exchanges = config.pop('exchange') if 'exchange' in config else {}
        queues = config.pop('queue') if 'queue' in config else {}

        if enforce:
            self.__dict__.update(config)
        else:
            self.__dict__.update((k, v) for k, v in config.items() if (k not in self.__dict__) or v)

        for exchange in exchanges:
            self.add_exchange(**exchange)
        for queue in queues:
            self.add_queue(**queue)

    def _check(self):
        if not self.host:
            raise ValueError("broker: invalid ip.")

        if not self.username:
            raise ValueError("broker: invalid user.")

        if not self.password:
            raise ValueError("broker: invalid passwd.")

    def _find_attr(self, keys: list, target: dict):
        res = {}
        for key in target:
            if key in keys:
                res[key] = target[key]
        return res

    @property
    def amqp(self):
        if self.host and self.username and self.password:
            return 'amqp://{}:{}@{}:{}/{}'.format(self.username, self.password, self.host, self.port, self.vhost)
        else:
            raise

    def connection(self):
        if not self._conn:
            self._check()
            self._conn = Connection(self.amqp, **self.kwargs)
        return self._conn

    def add_exchange(self, **kwarg):
        #: By default messages sent to exchanges are persistent (delivery_mode=2),
        #: and queues and exchanges are durable.
        _name = kwarg['name']
        _exchange = Exchange(**kwarg)
        self.exchanges[_name] = _exchange

    def add_queue(self, **kwarg):
        _name = kwarg['name']
        if kwarg['exchange'] in self.exchanges:
            kwarg['exchange'] = self.exchanges[kwarg['exchange']]
            _queue = Queue(**kwarg)
            self.queues[_name] = _queue
        else:
            logger.error('不存在 {} 这个exchange'.format(kwarg['exchange']))

    def add_conn(self, queue, routing_key, exchange_type, exchange, queue_attrs={}, exchange_attrs={}):
        _exchange = self.add_exchange(name=exchange, type=exchange_type, **exchange_attrs)
        _queue = self.add_queue(name=queue, exchange=exchange, routing_key=routing_key, **queue_attrs)
        self.exchanges[exchange] = _exchange
        self.queues[queue] = _queue
        return _queue, _exchange


class RabbitMQ:
    def __init__(self, broker, cb=None):
        self.broker = broker
        self.cb = cb if cb else self.handle_message
        self.is_running = False
        self.thread = None

    def run(self, queues=None, thread=True, **kwargs):
        res = []

        if not queues:
            res = [self.broker.queues[x] for x in self.broker.queues]
        else:
            if isinstance(queues, str):
                if queues in self.broker.queues:
                    res = [self.broker.queues[queues]]
                queues = [queues]
            for queue in set(queues):
                if queue not in self.broker.queues:
                    logger.error('no queue named {}'.format(queue))
                else:
                    res.append(self.broker.queues[queue])

        if res:
            if thread:
                self.thread = Thread(target=self._run_consumer, args=(res,), kwargs=kwargs)
                self.thread.start()
            else:
                self._run_consumer(res, **kwargs)
        else:
            logger.warning('no queue avaible, please check again!')
            return None

    @staticmethod
    def handle_message(body, message):
        #: This is the callback applied when a message is received.
        print(f'Received message: {body!r}')
        print(f'  properties:\n{pretty(message.properties)}')
        print(f'  delivery_info:\n{pretty(message.delivery_info)}')
        # finally should ack this message
        message.ack()

    def _run_consumer(self, queues, **kwargs):
        #: Create a connection and a channel.
        #: If hostname, userid, password and virtual_host is not specified
        #: the values below are the default, but listed here so it can
        #: be easily changed.
        timeout = kwargs.get('timeout') or 1
        ignore_timeouts = kwargs.get('ignore_timeouts') or True
        with self.broker.connection() as connection:
            with Consumer(connection, queues, callbacks=[self.cb], **kwargs):
                self.is_running = True
                logger.debug("mq is running at {}".format(self.broker.amqp))
                while self.is_running:
                    try:
                        # #: Each iteration waits for a single event.  Note that this
                        #         #: event may not be a message, or a message that is to be
                        #         #: delivered to the consumers channel, but any event received
                        #         #: on the connection.
                        for _ in eventloop(connection, timeout=timeout, ignore_timeouts=ignore_timeouts):
                            pass
                    except ConnectionForced:
                        self.is_running = False
                        logger.info('MQ Received a Force Close Commend: {}'.format(e))
                        break
                    except ConnectionResetError:
                        pass
                    except Exception as e:
                        self.is_running = False
                        logger.error('cb has some problem {}'.format(e), exc_info=True)
                        break

    def send_as_task(self, data, block=False, **kwargs):
        with producers[self.broker.connection()].acquire(block=block) as producer:
            producer.publish(data, **kwargs)

    # if __name__ == '__main__':
    #     broker = Broker()
    #     broker.load(config_path='some_file.yaml')
    #     mq = RabbitMQ(broker)
    #     mq.run(thread=False)
    #
    """
    demo some_file.yaml:
    MQCONFIG:
      host: 106.54.140.36
      port: 5672
      username: guest
      password: guest
      exchange:
        - name: message
          type: topic
          durable: false
        - name: asset.scanTaskTopic
          type: topic
          durable: true
      queue:
        - name: asset.scanTaskManage
          exchange: asset.scanTaskTopic
          routing_key: scanTaskManage
          durable: true
        - name: TEST01
          exchange: message
          routing_key: example.text
          durable: false
    
    """



import sqlite3
import logging

logger = logging.getLogger(__name__)


class Sqlite(object):

    def __init__(self, db_path):
        self.db = sqlite3.connect(db_path)

    def __del__(self):
        self.db.close()

    def execute_sql(self, sql):
        try:
            with self.db as conn:
                return conn.execute(sql)
        except Exception as e:
            logger.error('sql执行出错\nSQL: {} ,错误原因 {}'.format(sql, e), exc_info=True)
            return False



import sys
import time
import asyncio
import logging
from threading import Thread

logger = logging.getLogger(__name__)


# TODO 需要增加异常输出

class Coro(Thread):
    def __init__(self, name='coroThread', sem=40, daemon=False):
        """
        开一条线程单独去处理所有的协程任务
        :param name: 线程名
        :param sem: 协程并发量控制
        :param daemon:
        """
        super().__init__(name=name, daemon=daemon)
        self.loop = asyncio.new_event_loop()
        self.tasks = {}
        # self.loop.set_exception_handler(self.custom_exception_handler)
        self._sem_count = sem
        self.sem = None

    def __getitem__(self, item):
        try:
            return self.tasks[item]
        except KeyError:
            print('任务{}未存在'.format(item))
            raise

    def __repr__(self):
        d = {}
        for name in self.tasks:
            d[name] = {'state': self[name]._state}
        return str(d)

    def run(self) -> None:
        try:
            asyncio.set_event_loop(self.loop)
        except:
            logger.error('协程启动失败')
            sys.exit()
        try:
            self.loop.run_forever()
        finally:
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            self.loop.close()

    async def sem_task(self, coro):
        # 对task增加信号量，限制同一时间正在执行的任务数
        if not self.sem:
            self.sem = asyncio.Semaphore(self._sem_count)
        async with self.sem:
            return await coro

    def add_task(self, coro, name=None, sem=False):
        if name in self.tasks:
            return False
        if sem:
            task = self.loop.create_task(self.sem_task(coro))
        else:
            task = self.loop.create_task(coro, name=name)
        self.loop._csock.send(b'\0')
        name = task.get_name()
        self.tasks[name] = task
        return name

    def status(self, name):
        return self[name]._state

    def stop_task(self, name, nowait=False):
        self[name].cancel()
        if not nowait:
            while not self.tasks[name].cancelled():
                time.sleep(0.5)

    def get_result(self, name):
        try:
            res = self[name].result()
            del self.tasks[name]
        except asyncio.exceptions.CancelledError:
            print('Task {} 已取消'.format(name))
            res = '任务已取消'
        except Exception as e:
            res = str(e)
        return res

    def is_done(self, name):
        return self[name].done()



# coding: utf-8
import pymysql
from dbutils.pooled_db import PooledDB, SharedDBConnection

import logging

logger = logging.getLogger(__name__)


class Mysql(object):
    """
    注意：insert，update，delete操作时需要commit
    """

    def __init__(self, db, host, port, user, pwd):
        self.db = pymysql.connect(host=host, port=port, user=str(user), password=str(pwd))
        self.cursor = self.db.cursor()
        self.current_db = db

    def __del__(self):
        self.cursor.close()
        self.db.close()

    def execute_sql(self, sql):
        try:
            self.db.ping(reconnect=True)
            self.db.select_db(self.current_db)
            # sql = sql.lower()
            self.cursor.execute(sql)
            method = sql.split()[0]

            if method.startswith(('insert', 'update', 'delete')):
                self.db.commit()
                return True
            elif method.startswith('select'):
                return list(self.cursor.fetchall())
            else:
                self.db.commit()
                logger.error('未知sql类型'.format(sql))
                return False
        except Exception as e:
            logger.error('sql执行出错\nSQL: {} ,错误原因 {}'.format(sql, e), exc_info=True)
            return False

    def insert(self, tb, dt):
        """
        :param tb: 目标插入表
        :param dt: 字典型
        :return:
        """
        ls = [(k, dt[k]) for k in dt if dt[k] is not None]
        sql = 'insert %s (' % tb + ','.join(i[0] for i in ls) + \
              ') values (' + ','.join('%r' % i[1] for i in ls) + ');'
        self.execute_sql(sql)

    def select(self, tb, *columns, **factor):
        """
        :param tb: 目标插入表
        :param columns:select内容，空为全部
        :param factor: where内容

        example: self.select('arm_entry_services', asset_ip='192.168.90.26')
        """
        where = ''
        columns = '*' if columns == () or '' else ','.join(columns)

        if len(factor) > 0:
            where = 'where 1=1 '
            for column in factor:
                if factor[column] == '':
                    continue
                elif factor[column].startswith('like'):
                    conditional = 'and {} {}'.format(column, factor[column])
                else:
                    conditional = 'and {}={}'.format(column, "%r" % factor[column])
                if column == 'limit':
                    where += f'{column} {factor[column]}'
                else:
                    where += conditional + ' '

        sql = f'select {columns} from {tb} {where}'
        return self.execute_sql(sql)

    def update(self, tb, target_dic, set_dic):
        ts = [(k, target_dic[k]) for k in target_dic if target_dic[k] is not None]
        ss = [(k, set_dic[k]) for k in set_dic if set_dic[k] is not None]
        sql = 'UPDATE {} SET {} WHERE {}'.format(tb, ','.join([i[0] + '=' + '%r' % i[1] for i in ss]),
                                                 ' AND '.join([i[0] + '=' + '%r' % i[1] for i in ts]))
        self.execute_sql(sql)


class MysqlPool(object):
    """
    mysql连接池，用于处理多线程
    """

    def __init__(self, db, host, port, user, pwd, **kwargs):

        pool_config = {
            'creator': pymysql,  # 使用链接数据库的模块
            'maxconnections': 6,  # 连接池允许的最大连接数，0和None表示不限制连接数
            'mincached': 2,  # 初始化时，链接池中至少创建的空闲的链接，0表示不创建
            'maxcached': 5,  # 链接池中最多闲置的链接，0和None不限制
            'maxshared': 3,
            # 链接池中最多共享的链接数量，0和None表示全部共享。PS: 无用，因为pymysql和MySQLdb等模块的 threadsafety都为1，所有值无论设置为多少，_maxcached永远为0，所以永远是所有链接都共享。
            'blocking': True,  # 连接池中如果没有可用连接后，是否阻塞等待。True，等待；False，不等待然后报错
            'maxusage': None,  # 一个链接最多被重复使用的次数，None表示无限制
            'setsession': [],  # 开始会话前执行的命令列表。如：["set datestyle to ...", "set time zone ..."]
            'ping': 1,
            # ping MySQL服务端，检查是否服务可用。# 如：0 = None = never, 1 = default = whenever it is requested, 2 = when a cursor is created, 4 = when a query is executed, 7 = always
            'host': host,
            'port': port,
            'user': user,
            'password': pwd,
            'database': db,
            'charset': 'utf8'
        }
        pool_config.update(**kwargs)

        self.pool = PooledDB(**pool_config)

    def __new__(cls, *args, **kw):
        '''
        启用单例模式
        :param args:
        :param kw:
        :return:
        '''
        if not hasattr(cls, '_instance'):
            cls._instance = object.__new__(cls)
        return cls._instance

    def __del__(self):
        try:
            self.pool.close()
        except Exception:
            pass

    def execute_sql(self, sql, **args):
        try:
            sql = sql.lower()
            conn, cursor = self.connect()
            cursor.execute(sql, args)

            if sql.startswith(('insert', 'update', 'delete')):
                conn.commit()
                self.connect_close(conn, cursor)
                return True
            elif sql.startswith('select'):
                record_list = cursor.fetchall()
                self.connect_close(conn, cursor)
                return record_list
            else:
                conn.commit()
                # logger.error('未知sql类型'.format(sql))
                return True
        except Exception as e:
            logger.error('sql执行出错\nSQL: {} ,错误原因 {}'.format(sql, e), exc_info=True)
            return False

    def connect(self):
        '''
        启动连接
        :return:
        '''
        conn = self.pool.connection()
        cursor = conn.cursor(cursor=pymysql.cursors.DictCursor)
        return conn, cursor

    def connect_close(self, conn, cursor):
        '''
        关闭连接
        :param conn:
        :param cursor:
        :return:
        '''
        cursor.close()
        conn.close()

    def select_all(self, sql, args):
        '''
        批量查询
        :param sql:
        :param args:
        :return:
        '''
        conn, cursor = self.connect()

        cursor.execute(sql, args)
        record_list = cursor.fetchall()
        self.connect_close(conn, cursor)

        return record_list

    def select_one(self, sql, args):
        '''
        查询单条数据
        :param sql:
        :param args:
        :return:
        '''
        conn, cursor = self.connect()
        cursor.execute(sql, args)
        result = cursor.fetchone()
        self.connect_close(conn, cursor)

        return result

    def execute(self, sql, args):
        """
        执行insert/delete/update操作
        :param sql:
        :param args:
        :return:
        """
        conn, cursor = self.connect()
        row = cursor.execute(sql, args)
        conn.commit()
        self.connect_close(conn, cursor)
        return row


class TaskMysql(MysqlPool, Mysql):
    def __init__(self, db, host, port, user, pwd, **kwargs):
        super().__init__(db=str(db), host=host, port=port,
                         user=str(user), pwd=str(pwd), **kwargs)
